import unittest

import pipeline_with_complex_task_groups
import pipeline_with_dynamic_dag
import pipeline_with_duck_typing
from test_utils import TestSandboxDir


class DynamicDagTests(unittest.TestCase):

    def test_pipeline_with_dynamic_dag(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_with_dynamic_dag.pipeline_task_generator, completed=True
        )

        for t in pipeline_instance.tasks:
            self.assertEqual(t.get_state().state_name, "completed")

        agg_task = pipeline_instance.tasks["aggregate_all"]

        self.assertEqual(agg_task.out.aggregate_inflated_number.fetch(), 20)

        self.assertEqual(
            {s for s in agg_task.out.insane_strings_passed_through.fetch().split(",")},
            {s for s in "abc1,abc2,abc4,abc3".split(",")}
        )

    def test_pipeline_with_complex_task_groups(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_with_complex_task_groups.pipeline_task_generator, completed=True
        )

        agg_task = pipeline_instance.tasks["grand_total"]

        self.assertEqual(agg_task.out.grand_total.fetch(), 240)

    def _test_pipeline_with_duck_typing(self):

        d = TestSandboxDir(self)

        s = [45, 3, 365, 6, 6788, 34, 2, 111]

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_with_duck_typing.create_dag_generator(s),
            completed=True
        )

        for t in pipeline_instance.tasks:
            self.assertEqual(t.get_state().state_name, "completed")




import itertools
import os
import statistics

from dry_pipe import Task
from dry_pipe.task_state import TaskState, VALID_TRANSITIONS, STATE_EQUIVALENCE_FOR_DISPLAY, \
    STATE_EQUIVALENCE_FOR_DISPLAY_WEB


class Counter:

    def __init__(self):
        self.value = 0

    def inc(self):
        self.value += 1


class TaskGroupMetric:

    def __init__(self, task_group_key, display_order_key):
        self.count_per_state_name = {}
        self.task_group_key = task_group_key
        self.display_order_key = display_order_key

    def inc(self, state_name):
        counter = self.count_per_state_name.get(state_name)

        if counter is None:
            counter = Counter()
            self.count_per_state_name[state_name] = counter

        counter.inc()

    def counts(self, state_names_groups):

        def f(counter):
            if counter is None:
                return 0
            else:
                return counter.value

        def count_states_names(state_names):
            c = 0
            for state_name in state_names:
                c += f(self.count_per_state_name.get(state_name))

            return c

        return [
            count_states_names(state_names)
            for state_names in state_names_groups
        ]


class PipelineMetricsTable:

    @staticmethod
    def detailed_table_from_task_group_metrics(task_group_metrics):

        all_states = [
            s
            for s in VALID_TRANSITIONS.keys()
            if s != ""
        ]

        all_states_groups = [
            [s]
            for s in all_states
        ]

        return PipelineMetricsTable._table_from_task_group_metrics(
            all_states,
            all_states_groups,
            task_group_metrics
        )

    @staticmethod
    def summarized_table_from_task_group_metrics(task_group_metrics):
        return PipelineMetricsTable._table_from_task_group_metrics(
            TaskState.state_display_names(),
            TaskState.display_groups_of_state_names(),
            task_group_metrics
        )

    @staticmethod
    def _table_from_task_group_metrics(state_display_names, display_groups_of_state_names, task_group_metrics):

        header_row = ["task group"] + state_display_names + ["total"]

        count_totals = [
            0
            for n in display_groups_of_state_names
        ]

        table_body = []

        for task_group_metric in task_group_metrics:

            row = [task_group_metric.task_group_key]

            counts = task_group_metric.counts(display_groups_of_state_names)

            for i in range(0, len(count_totals)):
                count_totals[i] += counts[i]
                row.append(counts[i])

            row.append(sum(counts))

            table_body.append(row)

        footer = [None] + count_totals + [sum(count_totals)]

        return header_row, table_body, footer

    @staticmethod
    def totals_row(pipeline_instance_dir):

        res = {
            s: 0
            for s in TaskState.state_display_names_web
        }

        for task_state in TaskState.fetch_all(pipeline_instance_dir=pipeline_instance_dir):

            display_state = STATE_EQUIVALENCE_FOR_DISPLAY_WEB[task_state.state_name]

            res[display_state] += 1

        return res


def fetch_task_group_metrics(pipeline_instance_dir, task_key_grouper, task_state_visitor=None, task_filter=None):

    task_group_metrics = {}

    for task_state in TaskState.fetch_all(pipeline_instance_dir):

        if task_filter is not None:
            if not task_filter(task_state.task_key):
                continue

        if task_state_visitor is not None:
            task_state_visitor(task_state)

        task_group_key = task_key_grouper(task_state.task_key)

        m = task_group_metrics.get(task_group_key)

        if m is None:
            m = TaskGroupMetric(task_group_key, os.path.getctime(task_state.control_dir()))
            task_group_metrics[task_group_key] = m

        m.inc(task_state.state_name)

    return sorted(
        task_group_metrics.values(),
        key=lambda group: group.display_order_key
    )


def fetch_task_groups_stats(pipeline_instance_dir, no_header=False, units='seconds'):

    def gen_step_durations():

        for task_state in TaskState.fetch_all(pipeline_instance_dir):

            task_group_key = Task.key_grouper(task_state.task_key)

            def row_name(history_row):
                return history_row[0]

            def row_time(history_row):
                return TaskState.parse_history_timestamp(history_row[1])

            def row_step_number(history_row):
                return int(history_row[2])

            def step_intervals():
                rows_accumulation = []
                for history_row in task_state.load_history_rows():
                    if row_name(history_row) in ["step-started", "step-completed"]:
                        rows_accumulation.append(history_row)
                    elif row_name(history_row) in ["timed-out", "failed", "crashed"]:
                        rows_accumulation = []
                yield from rows_accumulation

            for step_number, rows, in itertools.groupby(
                    sorted(step_intervals(), key=row_step_number),
                    key=row_step_number
            ):
                rows = list(rows)
                if len(rows) < 2:
                    continue
                last_step_start_row = rows[-2]
                last_step_completion_row = rows[-1]

                if row_name(last_step_start_row) != "step-started":
                    continue

                if row_name(last_step_completion_row) != "step-completed":
                    continue

                total_time = (row_time(last_step_completion_row) - row_time(last_step_start_row)).total_seconds()
                yield f"{task_group_key}:{step_number}", total_time

    if not no_header:
        yield "task:step", "min_t", "max_t", "total_t", "avg_t", "s_dev"

    if units == "minutes":
        unit_converter = lambda i: round(i / 60, 2)
    elif units == "seconds":
        unit_converter = lambda i: round(i)
    elif units == 'hours':
        unit_converter = lambda i: round(i / (60 * 60), 2)
    else:
        raise Exception(f"unknown units: {units}, valid units are 'seconds', 'minutes', 'hours' ")

    for step_name, step_duration_tuples in itertools.groupby(
            sorted(gen_step_durations(), key=lambda t: t[0]),
            key=lambda t: t[0]
    ):

        step_durations = sorted([
            time for s, time in step_duration_tuples
        ])

        min_t = step_durations[0]
        max_t = step_durations[-1]
        total_t = sum(step_durations)
        avg_t = total_t / len(step_durations)
        s_dev = 0 if len(step_durations) == 1 else statistics.stdev(step_durations)

        yield (
            step_name,
            unit_converter(min_t),
            unit_converter(max_t),
            unit_converter(total_t),
            unit_converter(avg_t),
            s_dev
        )



import os.path
import unittest

from dry_pipe import DryPipe, TaskConf


class CliTests(unittest.TestCase):

    def test_prepare_remote_sites(self):

        def tasks(dsl):
            yield dsl.task(key="abc")\
                .calls("""
                    #!/usr/bin/env bash
                    echo "nothing.."
                """)

        this_dir = os.path.dirname(os.path.abspath(__file__))

        pipeline = DryPipe.create_pipeline(
            tasks,
            pipeline_code_dir=os.path.join(this_dir, "test_code_dir"),
            remote_task_confs=[
                TaskConf(
                    executer_type="process",
                    ssh_specs=f"maxl@ip29.ccs.usherbrooke.ca:~/.ssh/id_rsa",
                    remote_pipeline_code_dir="/home/maxl/test_prepare_remote_sites"
                )
            ]
        )
        pipeline.prepare_remote_sites()



import logging
import os

from dry_pipe import PortablePopen, TaskState
from dry_pipe.actions import TaskAction
from dry_pipe.pubsub import SubscriptionRegistry
from dry_pipe.pubsub_messages import all_pipeline_states_as_json, task_details_message, pipeline_counts_message, \
    full_instances_dir_for_pdir

logger = logging.getLogger(__name__)


class UIJanitor:

    def __init__(self, instance_dirs_to_pipelines):
        self.instance_dirs_to_pipelines = instance_dirs_to_pipelines

    def _pack_message(self, event, sids, observed_key, data):

        logger.debug("packing message %s for %s, data:%s", event, sids, data)

        return {
            "event": event,
            "sids": sids,
            "observed_key": observed_key,
            "data": data
        }

    def messages_for_round(self):

        sids = SubscriptionRegistry.instance.sids()

        if len(sids) > 0:
            yield self._pack_message(
                "running-pipelines", sids, "", all_pipeline_states_as_json(self.instance_dirs_to_pipelines)
            )

        for task_key, sids in SubscriptionRegistry.instance.task_keys_to_sids().items():
            yield self._pack_message(
                "latestTaskDetails", sids, task_key, task_details_message(self.instance_dirs_to_pipelines, task_key)
            )

        for pid, sids in SubscriptionRegistry.instance.pids_to_sids().items():
            yield self._pack_message(
                "latestPipelineDetailedState", sids, pid, pipeline_counts_message(self.instance_dirs_to_pipelines, pid)
            )

    def message_after_subscription_update(self, action_from_browser):

        if action_from_browser["name"] == "observePipeline":

            pid = action_from_browser["pid"]

            yield self._pack_message(
                "latestPipelineDetailedState",
                [action_from_browser["sid"]],
                pid,
                pipeline_counts_message(self.instance_dirs_to_pipelines, pid)
            )

        elif action_from_browser["name"] == "observeTask":

            task_key = action_from_browser["task_key"]

            yield self._pack_message(
                "latestTaskDetails",
                [action_from_browser["sid"]],
                task_key,
                task_details_message(self.instance_dirs_to_pipelines, task_key)
            )

    def task_action_submitted(self, sid, message):

        task_key = message['task_key']
        pipeline_dir = message['pipeline_dir']
        action_name = message['action_name']
        is_cancel = message.get("is_cancel")

        pdir, pid = pipeline_dir.split("|")

        instances_dir = full_instances_dir_for_pdir(self.instance_dirs_to_pipelines, pdir)

        pipeline_instance_dir = os.path.join(instances_dir, pid)
        task_control_dir = os.path.join(pipeline_instance_dir, ".drypipe", task_key)

        if action_name == "kill":
            cmd = [os.path.join(task_control_dir, "task"), "kill"]
            logger.info("will run %s", cmd)
            p = PortablePopen(cmd)
            p.wait_and_raise_if_non_zero()
            logger.info("\n".join(p.read_stdout_lines()))
        elif action_name == "restart":
            task_state = TaskState.from_task_control_dir(pipeline_instance_dir, task_key)
            task_state.transition_to_prepared(force=True)
            logger.info("task %s requeued", task_control_dir)

        elif is_cancel is not None and is_cancel:
            # load existing action:
            action = TaskAction.load_from_task_control_dir(
                os.path.join(pipeline_instance_dir, ".drypipe", task_key)
            )
            if action is not None:
                action.delete()
        else:
            TaskAction.submit(pipeline_instance_dir, task_key, action_name, message.get("step"))

        return self._pack_message(
            "taskActionSubmitted", [sid], task_key, message
        )



import os
import unittest

import pipeline_for_ssh_crash_recovery_tests
import pipeline_with_single_python_task
from dry_pipe import DryPipe, TaskConf
from dry_pipe.janitors import Janitor
from dry_pipe.pipeline import Pipeline
from pipeline_with_multistep_tasks import three_steps_pipeline_task_generator, three_steps_pipeline_expected_output
from test_01_simple_static_pipeline.simple_static_pipeline import simple_static_pipeline, validate_pipeline_execution
from test_utils import TestSandboxDir, copy_pre_existing_file_deps_from_code_dir, ensure_remote_dirs_dont_exist
import test_utils


class DaemonModeTests(unittest.TestCase):

    def test_load_instances_from_base_dir(self):

        d = TestSandboxDir(self)

        pipeline = DryPipe.create_pipeline(simple_static_pipeline)

        i = Pipeline.pipeline_instances_iterator({d.sandbox_dir: pipeline})

        self.assertEqual(len(list(i)), 0)

        i.create_instance_in(d.sandbox_dir, "p1")

        self.assertEqual(len(list(i)), 1)

        p1 = list(i)[0]

        self.assertTrue(p1.pipeline_instance_dir.endswith("p1"))

        i.create_instance_in(d.sandbox_dir, "p2")
        i.create_instance_in(d.sandbox_dir, "p3")

        self.assertEqual(len(list(i)), 3)

    def test_load_instances_two_types_from_base_dir(self):

        d = TestSandboxDir(self)
        d1 = d.create_subdir("three_steps_pipelines")
        d2 = d.create_subdir("pipelines_with_single_python_task")

        pipeline1 = DryPipe.create_pipeline(three_steps_pipeline_task_generator)
        pipeline2 = DryPipe.create_pipeline(pipeline_with_single_python_task.pipeline)

        i = pipeline1.pipeline_instances_iterator({
            d1: pipeline1,
            d2: pipeline2
        })

        self.assertEqual(len(list(i)), 0)

        i.create_instance_in(d1, "p1")

        self.assertEqual(len(list(i)), 1)

        p1 = list(i)[0]

        self.assertTrue(p1.pipeline_instance_dir.endswith("p1"))

        i.create_instance_in(d1, "p2")
        i.create_instance_in(d2, "p3")

        self.assertEqual(len(list(i)), 3)

    def test_run_two_instances_until_done(self):

        d = TestSandboxDir(self)

        pipeline = DryPipe.create_pipeline(simple_static_pipeline)

        i = Pipeline.pipeline_instances_iterator({d.sandbox_dir: pipeline})
        i.create_instance_in(d.sandbox_dir, "p1")
        i.create_instance_in(d.sandbox_dir, "p2")

        for i0 in i:
            copy_pre_existing_file_deps_from_code_dir(i0)

        janitor = Janitor(pipeline_instances_iterator=i)

        work_iterator = janitor.iterate_main_work(sync_mode=True)

        has_work = next(work_iterator)
        while has_work:
            has_work = next(work_iterator)

        c = 0

        for p in Pipeline.pipeline_instances_iterator(
                {d.sandbox_dir: pipeline}, ignore_completed=False
        ):
            validate_pipeline_execution(p, self)
            c += 1

        self.assertEqual(c, 2)

    def test_run_instances_two_types_from_base_dir_until_done(self):

        d = TestSandboxDir(self)
        d1 = d.create_subdir("three_steps_pipelines")
        d2 = d.create_subdir("pipelines_with_single_python_task")

        pipeline1 = DryPipe.create_pipeline(three_steps_pipeline_task_generator)
        pipeline2 = DryPipe.create_pipeline(pipeline_with_single_python_task.pipeline)

        dir_to_pipeline = {
            d1: pipeline1,
            d2: pipeline2
        }

        i = Pipeline.pipeline_instances_iterator(dir_to_pipeline)

        i.create_instance_in(d1, "p1_three_steps_pipelines")
        i.create_instance_in(d2, "p2_pipelines_with_single_python_task")

        janitor = Janitor(pipeline_instances_iterator=i)

        work_iterator = janitor.iterate_main_work(sync_mode=True)

        has_work = next(work_iterator)
        while has_work:
            has_work = next(work_iterator)

        c = 0

        for p in Pipeline.pipeline_instances_iterator(
            dir_to_pipeline, ignore_completed=False
        ):
            if p.instance_dir_base_name() == "p1_three_steps_pipelines":
                three_phase_task = p.tasks["three_phase_task"]
                self.assertEqual(three_steps_pipeline_expected_output(), three_phase_task.out.out_file.load_as_string())
            elif p.instance_dir_base_name() == "p2_pipelines_with_single_python_task":
                pipeline_with_single_python_task.validate_single_task_pipeline(p)
            else:
                raise Exception(f"unknown pipeline {p.instance_dir_base_name()}")
            c += 1

        self.assertEqual(c, 2)


class LongRunningDaemonModeTests(unittest.TestCase):

    def test_ssh_crash_recovery(self):

        test_utils.log_4_debug_daemon_mode()

        tc_remote = TaskConf(
            executer_type="process",
            #ssh_specs="maxl@127.0.0.1:~/.ssh/id_dsa",
            ssh_specs="maxl@ip29.ccs.usherbrooke.ca:~/.ssh/id_rsa",
            remote_base_dir="/home/maxl/drypipe_tests",
            remote_containers_dir="dummyZzz"
        )

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_for_ssh_crash_recovery_tests.create_pipeline(
                #TaskConf.default()
                tc_remote
            )
        )

        ensure_remote_dirs_dont_exist(pipeline_instance)

        janitor = Janitor(pipeline_instance=pipeline_instance)
        thread = janitor.start()
        janitor.start_remote_janitors()
        thread.join()

        # socat tcp-listen:7775,bind=127.0.0.1,fork tcp:maxl@ip29.ccs.usherbrooke.ca:22
        # sudo tc qdisc add dev wlp1s0 root netem loss 30%





import unittest

from dry_pipe.monitoring import TaskGroupMetric, PipelineMetricsTable


class MonitoringTests(unittest.TestCase):

    def test_counters(self):

        g = TaskGroupMetric("k", 3)

        g.inc("a")
        g.inc("a")
        g.inc("b")

        self.assertEqual([2, 1], g.counts([["a"], ["b"]]))

        self.assertEqual([2, 1, 0], g.counts([["a"], ["b"], ["c"]]))

        self.assertEqual([0, 3, 0], g.counts([["x"], ["a", "b"], ["c"]]))

    def test_metrics_gen(self):

        gA = TaskGroupMetric("A", 1)
        gB = TaskGroupMetric("B", 2)
        gC = TaskGroupMetric("A", 4)

        gA.inc("s1")

        gB.inc("s2")
        gB.inc("s2")

        gC.inc("s3")
        gC.inc("s3")
        gC.inc("s3")

        header, body, footer = PipelineMetricsTable._table_from_task_group_metrics(
            ["s1", "s2", "s3"],
            [["s1"], ["s2"], ["s3"]],
            [gA, gB, gC]
        )

        self.assertEqual(
            [['A', 1, 0, 0, 1],
             ['B', 0, 2, 0, 2],
             ['A', 0, 0, 3, 3]],
            body
        )

        self.assertEqual(
            [None, 1, 2, 3, 6],
            footer
        )

        self.assertEqual(
            ['task group', 's1', 's2', 's3', 'total'],
            header
        )






import logging
import os
import time

from dry_pipe import Task
from dry_pipe.actions import TaskAction
from dry_pipe.monitoring import PipelineMetricsTable, fetch_task_group_metrics
from dry_pipe.pipeline_state import PipelineState
from dry_pipe.task_state import TaskState

logger = logging.getLogger(__name__)

def all_pipeline_states_as_json(instance_dirs_to_pipelines):

    def g():
        for d, p in instance_dirs_to_pipelines.items():
            for pipeline_state in PipelineState.iterate_from_instances_dir(d):
                if pipeline_state.is_completed():
                    continue
                yield {
                    "dir": pipeline_state.instance_dir_key(),
                    "state": pipeline_state.state_name,
                    "totals": pipeline_state.totals_row()
                }

    return list(g())


def full_instances_dir_for_pdir(instance_dirs_to_pipelines, pdir):
    for d, p in instance_dirs_to_pipelines.items():
        last = os.path.split(d)[-1]
        if last == pdir:
            return d



def task_details_message(instance_dirs_to_pipelines, pid_task_key):

    pdir, pid, task_key = pid_task_key.split("|")
    instances_dir = full_instances_dir_for_pdir(instance_dirs_to_pipelines, pdir)

    task_state = TaskState.from_task_control_dir(
        os.path.join(instances_dir, pid), task_key
    )

    return task_state.as_json(timeout=5)


def pipeline_counts_message(instance_dirs_to_pipelines, pid):

    def prepare_counts_message(pipeline_instance_dir):

        pipeline_state = PipelineState.from_pipeline_instance_dir(pipeline_instance_dir)

        actions_by_task_key = {
            task_action.task_key: task_action
            for task_action in TaskAction.fetch_all_actions(pipeline_state.instance_dir())
        }

        def create_visitor():

            partial_task_infos = []

            def visit_task(task_state):

                o = {
                    "key": task_state.task_key,
                    "state_name": task_state.state_name,
                    "step": task_state.step_number()
                }

                err_tail = [] #task_state.tail_err_if_failed(3)

                if err_tail is not None:
                    o["err_tail"] = err_tail

                action = actions_by_task_key.get(task_state.task_key)

                if action is not None:
                    o["action"] = action.action_name

                partial_task_infos.append(o)

            return partial_task_infos, visit_task

        partial_task_infos, task_state_visitor = create_visitor()

        snapshot_time = int(time.time_ns())

        return {
            # SHOULD SEND DAG HERE ?
            # "dag": pipeline.summarized_dependency_graph(),
            "tsv": all_task_states_as_tsv(pipeline_state.instance_dir(), task_state_visitor),
            "pipelineDir": pipeline_state.instance_dir_key(),
            "partial_task_infos": partial_task_infos,
            "snapshot_time": snapshot_time
        }

    def all_task_states_as_tsv(pipeline_instance_dir, task_state_visitor=None):
        header, table_body, footer = PipelineMetricsTable.detailed_table_from_task_group_metrics(
            fetch_task_group_metrics(
                pipeline_instance_dir,
                Task.key_grouper,
                task_state_visitor=task_state_visitor
            )
        )

        def format_cell(c):
            return str(c or 0)

        res = "\n".join([
            "\t".join([
                format_cell(cell) for cell in row
            ])
            for row in table_body
        ])

        header = '\t'.join(header[1:-1])

        return f"{header}\n{res}"

    pdir, pid = pid.split("|")

    pipeline_instance_dir = os.path.join(
        full_instances_dir_for_pdir(instance_dirs_to_pipelines, pdir),
        pid
    )

    pipeline_state = PipelineState.from_pipeline_instance_dir(pipeline_instance_dir)

    return prepare_counts_message(pipeline_state.instance_dir())



import logging
import os
import traceback
from contextlib import contextmanager
from itertools import groupby
from timeit import default_timer
import requests

from dry_pipe.script_lib import PortablePopen


def send_email_error_report_if_configured(subject_line, exception=None, details=None):

    logger = logging.getLogger()

    error_email_config = os.environ.get("DRYPIPE_ERROR_EMAIL")

    if error_email_config is None:
        logger.info("DRYPIPE_ERROR_EMAIL not set")
        return

    mailgun_adresseses, mailgun_domain, mailgun_private_api_key, instance_label = error_email_config.split(":")

    logger.info("will send error report to %s with domain %s", mailgun_adresseses, mailgun_domain)

    if exception is not None:
        text = exception_to_string(exception)
    elif details is not None:
        text = details
    else:
        text = subject_line

    data = {
        "from": f"drypipe@{mailgun_domain}",
        "to": mailgun_adresseses.split(","),
        "subject": f"{subject_line}, {instance_label}",
        "text": f"""
            {text}
            INSTANCE: {instance_label}        
        """
    }

    requests.post(
        'https://api.mailgun.net/v2/{}/messages'.format(mailgun_domain),
        data=data,
        auth=('api', mailgun_private_api_key)
    )


def exception_to_string(exception):
    return "".join(traceback.TracebackException.from_exception(exception).format())


@contextmanager
def elapsed_timer():
    start = default_timer()
    elapser = lambda: default_timer() - start
    yield lambda: elapser()


logger_perf = logging.getLogger("dry_pipe.perf")

#   logger_perf.setLevel(logging.DEBUG)
#   logger_perf.addHandler(logging.StreamHandler(sys.stdout))

@contextmanager
def perf_logger_timer(group_label, details="", rounding_decimals=3, logger=logger_perf):

    if logger.level != logging.DEBUG:
        yield ""
        return

    start = default_timer()
    try:
        yield ""
    finally:
        logger.debug(
            "%s\t%s\t%s",
            group_label,
            round(default_timer() - start, rounding_decimals),
            details
        )

def analyze_perf_log(log_file, rounding_decimals=2):

    def enum_rows():
        with open(log_file) as f:
            for line in f.readlines():
                yield line.split("\t")

    def call_label(row):
        return row[0]

    def time_taken(row):
        return float(row[1])

    def gen_stats():
        for label, rows in groupby(sorted(enum_rows(), key=call_label), key=call_label):
            durations = list(map(time_taken, rows))
            s = sum(durations)
            avg = s / len(durations)
            yield round(avg, rounding_decimals), round(s, rounding_decimals), label

    for stat in gen_stats():
        stat = [
            str(s) for s in stat
        ]
        print("\t".join(stat))


def bash_shebang():
    return "#!/usr/bin/env bash"


def count_cpus():
    with PortablePopen(["grep", "-c", "^processor", "/proc/cpuinfo"]) as p:
        p.wait_and_raise_if_non_zero()
        return int(p.stdout_as_string())



import asyncio
import json
import logging
from threading import Thread

import socketio
import uvicorn
from socketio import AsyncNamespace

from dry_pipe.janitors import DaemonThreadHelper
from dry_pipe.pubsub import SubscriptionRegistry, DRYPIPE_WEBSOCKET_NAMESPACE
from dry_pipe.ui_janitor import UIJanitor

logger = logging.getLogger(__name__)


class WebsocketServer(AsyncNamespace):

    @staticmethod
    def start(bind_address, port, instance_dirs_to_pipelines):

        try:
            SubscriptionRegistry.init()

            server_websocket_adapter = WebsocketServer(instance_dirs_to_pipelines)

            def _is_port_in_use(bind_address, port):
                import socket
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    return s.connect_ex((bind_address, port)) == 0

            if _is_port_in_use(bind_address, port):
                logger.error(f" port {port} is used by another application, use --port=X to listen on another port")
                return

            logger.info("listening on %s", f"http://{bind_address}:{port}")

            socket_io_server = socketio.AsyncServer(async_mode='asgi')

            socket_io_server.register_namespace(server_websocket_adapter)

            janitor_thread = Thread(target=lambda: server_websocket_adapter.janitor())
            janitor_thread.start()
            logger.info("ui_janitor thread started")

            app = socketio.ASGIApp(
                socket_io_server,
                socketio_path="/socket.io",
                #other_asgi_app=App(),
                #    static_files={
                #        "/main.js": map_static_file("dist/main.js"),
                #        "/main.js.map": map_static_file("dist/main.js.map")
                #    }
            )

            logger.info(f"web monitor running on: http://{bind_address}:{port} (Press CTRL+C to quit)")

            uvicorn.run(
                #"dry_pipe.server:app",
                app,
                host=bind_address, port=port,
                log_level="error"
                #reload=True,
                #reload_dirs=["/home/maxou/dev/dry_pipe/dry_pipe"],
                #reload_delay=2.0
            )
        except Exception as ex:
            logger.exception(ex)

    def __init__(self, instance_dirs_to_pipelines):
        super(AsyncNamespace, self).__init__(DRYPIPE_WEBSOCKET_NAMESPACE)

        if instance_dirs_to_pipelines is None:
            raise Exception(f"instance_dirs_to_pipelines can't be None")

        self.drypipe_ui_session_id = None

        logger.info("will run local ui_janitor on dir %s", instance_dirs_to_pipelines)
        self.ui_janitor = UIJanitor(instance_dirs_to_pipelines)

    def janitor(self):

        def call_async(f):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            async def asyncify():
                await f()

            loop.run_until_complete(asyncify())
            loop.close()

        try:

            daemon_thread_helper = DaemonThreadHelper(logger, min_sleep=4, max_sleep=8, pipelines=None)

            while True:

                for m in self.ui_janitor.messages_for_round():

                    logger.debug(f"will send %s", m)
                    async def go():
                        await self.trigger_event("broadcast", None, m)

                    call_async(go)

                    daemon_thread_helper.register_work()

                daemon_thread_helper.end_round()

        except Exception as ex:
            logger.exception(ex)
            raise ex

    async def send_latest_subscription_registry(self):
        self.server.emit("latest_subscription_registry")

    def _authenticate(self, http_env):

        headers = http_env["asgi.scope"]["headers"]

        def auth_cookie():
            for h in headers:
                name, val = h
                name = name.decode("utf-8").lower()
                if name == "set-cookie":
                    cookies = val.decode("utf-8").split("=")[1]

    async def trigger_event(self, event, *args):
        try:
            await self._trigger_event(event, *args)
        except Exception as ex:
            logger.exception("trigger_event failed: %s %s", event, json.dumps(args))
            raise ex

    async def _trigger_event(self, event, *args):

        sid = args[0]

        #if event == "connect":
        #    # self._autheticate(args[1])
        #    return

        logger.debug("received event %s, %s", event, sid)

        if event == "announce_drypipe_ui":

            if self.ui_janitor is not None:
                logger.warning(
                    "this drypipe ui hub already has a local server, will not accept connections from ui servers"
                )
                return

            self.drypipe_ui_session_id = sid

            logger.info(f"ui server connected %s, will send reset_subscriptions message", self.drypipe_ui_session_id)

            await self.server.emit(
                "reset_subscriptions",
                {"sids_to_subscriptions": SubscriptionRegistry.instance.sids_to_subscriptions()},
                to=self.drypipe_ui_session_id,
                namespace=DRYPIPE_WEBSOCKET_NAMESPACE
            )

            #self.server.start_background_task(target=self.work)

        elif event == "update_subscription":

            action_from_browser = args[1]
            updated_subscription = SubscriptionRegistry.instance.update_subscription(action_from_browser)

            if self.ui_janitor is None:

                await self.server.emit(
                    "update_subscription",
                    {
                        "sids_to_subscriptions": SubscriptionRegistry.instance.sids_to_subscriptions(),
                        "action_from_browser": action_from_browser
                    },
                    to=self.drypipe_ui_session_id,
                    namespace=DRYPIPE_WEBSOCKET_NAMESPACE
                )
            else:

                for m in self.ui_janitor.message_after_subscription_update(action_from_browser):
                    await self.server.emit(m["event"], m["data"], sid=sid, namespace=DRYPIPE_WEBSOCKET_NAMESPACE)

            # sending the ack *after* calling drypipe_ui, ensures the user knows if the server is down

            await self.server.emit(
                "subscription_update_ack",
                {
                    "updated_subscription": updated_subscription,
                    "action_name": action_from_browser["name"]
                },
                to=sid,
                namespace=DRYPIPE_WEBSOCKET_NAMESPACE
            )

        elif event == "disconnect":
            if sid == self.drypipe_ui_session_id:
                logger.info(f"drypipe_ui_session_id %s disconnected", sid)
                self.drypipe_ui_session_id = None
            else:

                SubscriptionRegistry.instance.update_subscription({"name": "close", "sid": sid})

                await self.server.emit(
                    "reset_subscriptions",
                    {"sids_to_subscriptions": SubscriptionRegistry.instance.sids_to_subscriptions()},
                    to=self.drypipe_ui_session_id,
                    namespace=DRYPIPE_WEBSOCKET_NAMESPACE
                )

        elif event == "submitTaskAction":

            sid, message = args
            message["sid"] = sid

            logger.info("task submitted %s", message)

            if self.ui_janitor is not None:
                m = self.ui_janitor.task_action_submitted(sid, message)
                await self.trigger_event("broadcast", None, m)
            else:
                await self.server.emit(
                    "submitTaskAction",
                    message,
                    to=self.drypipe_ui_session_id,
                    namespace=DRYPIPE_WEBSOCKET_NAMESPACE
                )

        elif event == "broadcast":

            _sid_of_drypipe_ui_server_or_none, message = args

            relayed_event = message["event"]
            sids = message["sids"]

            logger.debug("will broadcast message %s to sids %s", relayed_event, sids)

            broadcast_to_all_failed_sids = []

            for sid in sids:
                try:
                    await self.server.emit(
                        relayed_event,
                        message["data"],
                        to=sid,
                        namespace=DRYPIPE_WEBSOCKET_NAMESPACE
                    )
                except Exception as exc:
                    logger.warning("failed while relaying event %s to %s\n%s", event, sid, exc)
                    broadcast_to_all_failed_sids.append(sid)

            for failed_sid in broadcast_to_all_failed_sids:
                logger.info("will disconnect %s", failed_sid)
                await self.server.disconnect(failed_sid)

            if len(broadcast_to_all_failed_sids) > 0:
                SubscriptionRegistry.instance.close_sessions(broadcast_to_all_failed_sids)

            #TODO: use this:
            #exceptions = await asyncio.gather(broadcast_to_all())


class App:
    async def __call__(self, scope, receive, send):
        if scope["type"] == "lifespan":
            return


if __name__ == '__main__':

    from dry_pipe.internals import setup_debug_log_config

    extra = {
        'socketio': {
            'handlers': ['console'],
            'level': "DEBUG",
            'propagate': False
        },
    }

    setup_debug_log_config(drypype_level="DEBUG", root_level="INFO", main_level="DEBUG", extra_loggers=extra)

    WebsocketServer.start("0.0.0.0", 5000)



import os
import unittest

from dry_pipe import TaskConf
from pipeline_with_never_ending_task import never_ending_pipeline
from pipelines_with_bash_corner_cases import dag_gen_for_set_e_test
from tests import pipeline_with_single_bash_task
from test_utils import TestSandboxDir, ensure_remote_dirs_dont_exist


class CornerCasesFailureTests(unittest.TestCase):

    def test_slurm_timeout_signal_handling(self):

        pipeline = TestSandboxDir(self).pipeline_instance_from_generator(
            never_ending_pipeline(TaskConf(
                executer_type="slurm",
                slurm_account="def-rodrigu1",
                sbatch_options=[
                    "--time=0:1:00"
                ]
            ))
        )

        single_task, = pipeline.tasks

        pipeline.clean()

        pipeline.run_sync()

        self.assertTrue(single_task.get_state().is_timed_out())

    def test_bash_set_e(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            dag_gen_for_set_e_test,
            completed=True
        )

        self.assertIsNone(pipeline_instance.tasks["t1"].out.r.fetch())



class CornerCasesRemoteZombiTests(unittest.TestCase):

    def test_slurm_zombi_detection(self):

        try:
            os.environ["IS_ZOMBI_TEST_CASE"] = "True"

            pipeline_instance = TestSandboxDir(self).pipeline_instance_from_generator(
                lambda dsl: pipeline_with_single_bash_task.pipeline(dsl, TaskConf(
                    executer_type="slurm",
                    slurm_account="def-rodrigu1",
                    sbatch_options=["--time=0:1:00"],
                    ssh_specs="maxl@ip29.ccs.usherbrooke.ca:~/.ssh/id_rsa",
                    remote_base_dir="/home/maxl/drypipe_tests"
                ))
            )

            ensure_remote_dirs_dont_exist(pipeline_instance)

            single_task, = pipeline_instance.tasks

            pipeline_instance.clean()

            pipeline_instance.run_sync()

            self.assertTrue(single_task.get_state().is_crashed())
        finally:
            del os.environ["IS_ZOMBI_TEST_CASE"]


from setuptools import setup

minimal_requirements = [
    'click==8.1.3',
    'uvicorn==0.17.6',
    'python-socketio==5.6.0',
    'rich==12.6.0',
    'requests==2.27.1',
    'readchar==4.0.3',
    'parallel-ssh==2.12.0'
]


setup(
    name='dry_pipe',
    version='0.7.51',
    packages=['dry_pipe'],
    install_requires=minimal_requirements,
    include_package_data=True,
    entry_points='''
        [console_scripts]
        drypipe=dry_pipe.cli:run_cli
    ''',
)